import napari.utils.notifications
from napari._qt.qt_event_loop import run
from napari.plugins.io import save_layers
from napari.view_layers import (
    view_image,
    view_labels,
    view_path,
    view_points,
    view_shapes,
    view_surface,
    view_tracks,
    view_vectors,
)
from napari.viewer import Viewer, current_viewer

__version__: str

notification_manager: napari.utils.notifications.NotificationManager

__all__ = (
    'Viewer',
    '__version__',
    'current_viewer',
    'notification_manager',
    'run',
    'save_layers',
    'view_image',
    'view_labels',
    'view_path',
    'view_points',
    'view_shapes',
    'view_surface',
    'view_tracks',
    'view_vectors',
)
